# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.repository"""

import logging
import time
import unittest

from logilab.database import get_db_helper
from yams import register_base_type, ValidationError, unregister_base_type
from yams.constraints import UniqueConstraint

from cubicweb import (
    UnknownEid,
    AuthenticationError,
    Unauthorized,
    QueryError,
)
from cubicweb.devtools.repotest import tuplify
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.predicates import is_instance
from cubicweb.schema import RQLConstraint
from cubicweb.server import hook
from cubicweb.server.hook import Hook
from cubicweb.server.sources import native
from cubicweb.server.sqlutils import SQL_PREFIX


class RepositoryTC(CubicWebTC):
    """singleton providing access to a persistent storage for entities
    and relation
    """

    def test_unique_together_constraint(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT Societe S: S nom "Logilab", S type "SSLL", S cp "75013"'
            )
            with self.assertRaises(ValidationError) as wraperr:
                cnx.execute(
                    'INSERT Societe S: S nom "Logilab", S type "SSLL", S cp "75013"'
                )
            self.assertEqual(
                {
                    "cp": "%(KEY-rtype)s is part of violated unicity constraint",
                    "nom": "%(KEY-rtype)s is part of violated unicity constraint",
                    "type": "%(KEY-rtype)s is part of violated unicity constraint",
                    "": "some relations violate a unicity constraint",
                },
                wraperr.exception.args[1],
            )

    def test_unique_together_schema(self):
        person = self.repo.schema.entity_schema_for("Personne")
        self.assertEqual(len(person._unique_together), 1)
        self.assertCountEqual(person._unique_together[0], ("nom", "prenom", "inline2"))

    def test_all_entities_have_owner(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertFalse(cnx.execute("Any X WHERE NOT X owned_by U"))

    def test_all_entities_have_is(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertFalse(cnx.execute("Any X WHERE NOT X is ET"))

    def test_all_entities_have_cw_source(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertFalse(cnx.execute("Any X WHERE NOT X cw_source S"))

    def test_connect(self):
        with self.repo.internal_cnx() as cnx:
            self.assertTrue(
                self.repo.authenticate_user(
                    cnx, self.admlogin, password=self.admpassword
                )
            )
            self.assertRaises(
                AuthenticationError,
                self.repo.authenticate_user,
                cnx,
                self.admlogin,
                password="nimportnawak",
            )
            self.assertRaises(
                AuthenticationError,
                self.repo.authenticate_user,
                cnx,
                self.admlogin,
                password="",
            )
            self.assertRaises(
                AuthenticationError,
                self.repo.authenticate_user,
                cnx,
                self.admlogin,
                password=None,
            )
            self.assertRaises(
                AuthenticationError,
                self.repo.authenticate_user,
                cnx,
                None,
                password=None,
            )
            self.assertRaises(
                AuthenticationError, self.repo.authenticate_user, cnx, self.admlogin
            )
            self.assertRaises(
                AuthenticationError, self.repo.authenticate_user, cnx, None
            )

    def test_login_upassword_accent(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                "INSERT CWUser X: X login %(login)s, X upassword %(passwd)s, "
                'X in_group G WHERE G name "users"',
                {"login": "barnabé", "passwd": "héhéhé".encode()},
            )
            cnx.commit()
            repo = self.repo
            self.assertTrue(
                repo.authenticate_user(cnx, "barnabé", password="héhéhé".encode())
            )

    def test_rollback_on_execute_validation_error(self):
        class ValidationErrorAfterHook(Hook):
            __regid__ = "valerror-after-hook"
            __select__ = Hook.__select__ & is_instance("CWGroup")
            events = ("after_update_entity",)

            def __call__(self):
                raise ValidationError(self.entity.eid, {})

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(ValidationErrorAfterHook):
                self.assertRaises(
                    ValidationError,
                    cnx.execute,
                    'SET X name "toto" WHERE X is CWGroup, X name "guests"',
                )
                self.assertTrue(cnx.execute('Any X WHERE X is CWGroup, X name "toto"'))
                with self.assertRaises(QueryError) as cm:
                    cnx.commit()
                self.assertEqual(str(cm.exception), "transaction must be rolled back")
                cnx.rollback()
                self.assertFalse(cnx.execute('Any X WHERE X is CWGroup, X name "toto"'))

    def test_rollback_on_execute_unauthorized(self):
        class UnauthorizedAfterHook(Hook):
            __regid__ = "unauthorized-after-hook"
            __select__ = Hook.__select__ & is_instance("CWGroup")
            events = ("after_update_entity",)

            def __call__(self):
                raise Unauthorized()

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(UnauthorizedAfterHook):
                self.assertRaises(
                    Unauthorized,
                    cnx.execute,
                    'SET X name "toto" WHERE X is CWGroup, X name "guests"',
                )
                self.assertTrue(cnx.execute('Any X WHERE X is CWGroup, X name "toto"'))
                with self.assertRaises(QueryError) as cm:
                    cnx.commit()
                self.assertEqual(str(cm.exception), "transaction must be rolled back")
                cnx.rollback()
                self.assertFalse(cnx.execute('Any X WHERE X is CWGroup, X name "toto"'))

    def test_initial_schema(self):
        schema = self.repo.schema
        # check order of attributes is respected
        notin = {
            "eid",
            "is",
            "is_instance_of",
            "identity",
            "creation_date",
            "modification_date",
            "cwuri",
            "owned_by",
            "created_by",
            "cw_source",
            "update_permission",
            "read_permission",
            "add_permission",
            "in_basket",
        }
        self.assertListEqual(
            [
                "relation_type",
                "from_entity",
                "to_entity",
                "constrained_by",
                "cardinality",
                "ordernum",
                "formula",
                "indexed",
                "fulltextindexed",
                "internationalizable",
                "defaultval",
                "extra_props",
                "description",
                "description_format",
            ],
            [
                r.type
                for r in schema.entity_schema_for("CWAttribute").ordered_relations()
                if r.type not in notin
            ],
        )

        self.assertEqual(schema.entity_schema_for("CWEType").main_attribute(), "name")
        self.assertEqual(schema.entity_schema_for("State").main_attribute(), "name")

        constraints = (
            schema.relation_schema_for("name")
            .relation_definition("CWEType", "String")
            .constraints
        )
        self.assertEqual(len(constraints), 2)
        for cstr in constraints[:]:
            if isinstance(cstr, UniqueConstraint):
                constraints.remove(cstr)
                break
        else:
            self.fail("unique constraint not found")
        sizeconstraint = constraints[0]
        self.assertEqual(sizeconstraint.min, None)
        self.assertEqual(sizeconstraint.max, 64)

        constraints = (
            schema.relation_schema_for("relation_type")
            .relation_definition("CWAttribute", "CWRType")
            .constraints
        )
        self.assertEqual(len(constraints), 1)
        cstr = constraints[0]
        self.assertIsInstance(cstr, RQLConstraint)
        self.assertEqual(cstr.expression, "O final TRUE")

        ownedby = schema.relation_schema_for("owned_by")
        self.assertEqual(ownedby.objects("CWEType"), ("CWUser",))

    def test_public_api(self):
        self.assertEqual(self.repo.get_schema(), self.repo.schema)
        self.assertEqual(
            self.repo.source_defs(),
            {"system": {"type": "native", "uri": "system", "use-cwuri-as-url": False}},
        )
        # .properties() return a result set
        self.assertEqual(
            self.repo.properties().rql,
            "Any K,V WHERE P is CWProperty,P pkey K, P value V, NOT P for_user U",
        )

    def test_schema_is_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            no_is_rset = cnx.execute("Any X WHERE NOT X is ET")
            self.assertFalse(no_is_rset, no_is_rset.description)

    def test_delete_if_singlecard1(self):
        with self.admin_access.repo_cnx() as cnx:
            note = cnx.create_entity("Affaire")
            p1 = cnx.create_entity("Personne", nom="toto")
            cnx.execute(
                "SET A todo_by P WHERE A eid %(x)s, P eid %(p)s",
                {"x": note.eid, "p": p1.eid},
            )
            rset = cnx.execute("Any P WHERE A todo_by P, A eid %(x)s", {"x": note.eid})
            self.assertEqual(len(rset), 1)
            p2 = cnx.create_entity("Personne", nom="tutu")
            cnx.execute(
                "SET A todo_by P WHERE A eid %(x)s, P eid %(p)s",
                {"x": note.eid, "p": p2.eid},
            )
            rset = cnx.execute("Any P WHERE A todo_by P, A eid %(x)s", {"x": note.eid})
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.rows[0][0], p2.eid)

    def test_delete_if_object_inlined_singlecard(self):
        with self.admin_access.repo_cnx() as cnx:
            c = cnx.create_entity("Card", title="Carte")
            cnx.create_entity("Personne", nom="Vincent", fiche=c)
            cnx.create_entity("Personne", nom="Florent", fiche=c)
            cnx.commit()
            self.assertEqual(len(c.reverse_fiche), 1)

    def test_delete_computed_relation_nonregr(self):
        with self.admin_access.repo_cnx() as cnx:
            c = cnx.create_entity("Personne", nom="Adam", login_user=cnx.user.eid)
            cnx.commit()
            c.cw_delete()
            cnx.commit()

    def test_cw_set_in_before_update(self):
        # local hook
        class DummyBeforeHook(Hook):
            __regid__ = "dummy-before-hook"
            __select__ = Hook.__select__ & is_instance("EmailAddress")
            events = ("before_update_entity",)

            def __call__(self):
                # safety belt: avoid potential infinite recursion if the test
                #              fails (i.e. RuntimeError not raised)
                pendings = self._cw.transaction_data.setdefault("pending", set())
                if self.entity.eid not in pendings:
                    pendings.add(self.entity.eid)
                    self.entity.cw_set(alias="foo")

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(DummyBeforeHook):
                addr = cnx.create_entity("EmailAddress", address="a@b.fr")
                addr.cw_set(address="a@b.com")
                rset = cnx.execute(
                    "Any A,AA WHERE X eid %(x)s, X address A, X alias AA",
                    {"x": addr.eid},
                )
                self.assertEqual(rset.rows, [["a@b.com", "foo"]])

    def test_cw_set_in_before_add(self):
        # local hook
        class DummyBeforeHook(Hook):
            __regid__ = "dummy-before-hook"
            __select__ = Hook.__select__ & is_instance("EmailAddress")
            events = ("before_add_entity",)

            def __call__(self):
                # cw_set is forbidden within before_add_entity()
                self.entity.cw_set(alias="foo")

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(DummyBeforeHook):
                # XXX will fail with python -O
                self.assertRaises(
                    AssertionError, cnx.create_entity, "EmailAddress", address="a@b.fr"
                )

    def test_multiple_edit_cw_set(self):
        """make sure cw_edited doesn't get cluttered
        by previous entities on multiple set
        """

        # local hook
        class DummyBeforeHook(Hook):
            _test = self  # keep reference to test instance
            __regid__ = "dummy-before-hook"
            __select__ = Hook.__select__ & is_instance("Affaire")
            events = ("before_update_entity",)

            def __call__(self):
                # invoiced attribute shouldn't be considered "edited" before the hook
                self._test.assertFalse(
                    "invoiced" in self.entity.cw_edited,
                    "cw_edited cluttered by previous update",
                )
                self.entity.cw_edited["invoiced"] = 10

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(DummyBeforeHook):
                cnx.create_entity("Affaire", ref="AFF01")
                cnx.create_entity("Affaire", ref="AFF02")
                cnx.execute("SET A duration 10 WHERE A is Affaire")

    def test_user_friendly_error(self):
        from cubicweb.entities.adapters import IUserFriendlyUniqueTogether

        class MyIUserFriendlyUniqueTogether(IUserFriendlyUniqueTogether):
            __select__ = IUserFriendlyUniqueTogether.__select__ & is_instance("Societe")

            def raise_user_exception(self):
                raise ValidationError(self.entity.eid, {"hip": "hop"})

        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(MyIUserFriendlyUniqueTogether):
                s = cnx.create_entity("Societe", nom="Logilab", type="ssll", cp="75013")
                cnx.commit()
                with self.assertRaises(ValidationError) as cm:
                    cnx.create_entity("Societe", nom="Logilab", type="ssll", cp="75013")
                self.assertEqual(cm.exception.errors, {"hip": "hop"})
                cnx.rollback()
                cnx.create_entity("Societe", nom="Logilab", type="ssll", cp="31400")
                with self.assertRaises(ValidationError) as cm:
                    s.cw_set(cp="31400")
                self.assertEqual(cm.exception.entity, s.eid)
                self.assertEqual(cm.exception.errors, {"hip": "hop"})
                cnx.rollback()

    def test_attribute_cache(self):
        with self.admin_access.repo_cnx() as cnx:
            bk = cnx.create_entity("Bookmark", title="index", path="/")
            cnx.commit()
            self.assertEqual(bk.title, "index")
            bk.cw_set(title="root")
            self.assertEqual(bk.title, "root")
            cnx.commit()
            self.assertEqual(bk.title, "root")

    def test_delete_entity_with_inlined_relation(self):
        """Test deletion of entity with inlined relations.

        In this case, inlined relation column should not be deleted (= e.g. not
        updated to NULL), but hooks before_delete_relation and
        afer_delete_relation must be called.
        """

        class OnDeleteInlined(Hook):
            __regid__ = "on_delete_inlined"
            __select__ = Hook.__select__ & hook.match_rtype("personne_inlined")
            events = ("before_delete_relation", "after_delete_relation")
            CALLED = []

            def __call__(self):
                OnDeleteInlined.CALLED.append(self.event)

        with self.admin_access.cnx() as cnx:
            # allow only one null on cw_personne_inlined column
            cnx.system_sql(
                "CREATE UNIQUE INDEX test_composite_idx ON cw_personne(true) "
                "WHERE cw_personne_inlined is NULL"
            )
            cnx.commit()

            # deletion of p1 should not set personne_inlined to NULL (otherwise
            # the unique index will raise)
            p0 = cnx.create_entity("Personne", nom="p0")
            p1 = cnx.create_entity("Personne", nom="p1", personne_inlined=p0)
            cnx.commit()
            with self.temporary_appobjects(OnDeleteInlined):
                cnx.entity_from_eid(p1.eid).cw_delete()
            assert OnDeleteInlined.CALLED == [
                "before_delete_relation",
                "after_delete_relation",
            ]
            cnx.commit()

            # XXX: This case is not handled because entities need to be deleted
            # in a specific order
            # p1 = cnx.create_entity('Personne', nom=u'p1', personne_inlined=p0)
            # cnx.commit()
            # cnx.execute('DELETE Personne X WHERE X eid IN ({}, {})'.format(p1.eid, p0.eid))
            # cnx.commit()
            cnx.entity_from_eid(p0.eid).cw_delete()
            cnx.commit()

            # deletion of p1 should not set personne_inlined to NULL
            p1 = cnx.create_entity("Personne", nom="p1")
            p1.cw_set(personne_inlined=p1)
            p0 = cnx.create_entity("Personne", nom="p0")
            cnx.commit()
            cnx.entity_from_eid(p1.eid).cw_delete()
            cnx.commit()


class SchemaDeserialTC(CubicWebTC):
    appid = "data-schemaserial"

    @classmethod
    def setUpClass(cls):
        register_base_type("BabarTestType", ("jungle_speed",))
        helper = get_db_helper("sqlite")
        helper.TYPE_MAPPING["BabarTestType"] = "TEXT"
        helper.TYPE_CONVERTERS["BabarTestType"] = lambda x: f'"{x}"'
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        unregister_base_type("BabarTestType")
        helper = get_db_helper("sqlite")
        helper.TYPE_MAPPING.pop("BabarTestType", None)
        helper.TYPE_CONVERTERS.pop("BabarTestType", None)
        super().tearDownClass()

    def test_deserialization_base(self):
        """Check the following deserialization

        * all CWEtype has name
        * Final type
        * CWUniqueTogetherConstraint
        * _unique_together__ content"""
        origshema = self.repo.schema
        try:
            self.repo.config.repairing = True  # avoid versions checking
            self.repo.set_schema(self.repo.deserialize_schema())
            table = SQL_PREFIX + "CWEType"
            namecol = SQL_PREFIX + "name"
            finalcol = SQL_PREFIX + "final"
            with self.admin_access.repo_cnx() as cnx:
                cu = cnx.system_sql(
                    f"SELECT {namecol} FROM {table} WHERE {finalcol} is NULL"
                )
                self.assertEqual(cu.fetchall(), [])
                cu = cnx.system_sql(
                    "SELECT %s FROM %s "
                    "WHERE %s=%%(final)s ORDER BY %s"
                    % (namecol, table, finalcol, namecol),
                    {"final": True},
                )
                self.assertEqual(
                    cu.fetchall(),
                    [
                        ("BabarTestType",),
                        ("BigInt",),
                        ("Boolean",),
                        ("Bytes",),
                        ("Date",),
                        ("Datetime",),
                        ("Decimal",),
                        ("Float",),
                        ("Int",),
                        ("Interval",),
                        ("Password",),
                        ("String",),
                        ("TZDatetime",),
                        ("TZTime",),
                        ("Time",),
                    ],
                )
                sql = (
                    "SELECT etype.cw_eid, etype.cw_name, cstr.cw_eid, rel.eid_to "
                    "FROM cw_CWUniqueTogetherConstraint as cstr, "
                    "     relations_relation as rel, "
                    "     cw_CWEType as etype "
                    "WHERE cstr.cw_eid = rel.eid_from "
                    "  AND cstr.cw_constraint_of = etype.cw_eid "
                    "  AND etype.cw_name = 'Personne' "
                    ";"
                )
                cu = cnx.system_sql(sql)
                rows = cu.fetchall()
                self.assertEqual(len(rows), 3)
                person = self.repo.schema.entity_schema_for("Personne")
                self.assertEqual(len(person._unique_together), 1)
                self.assertCountEqual(
                    person._unique_together[0], ("nom", "prenom", "inline2")
                )

        finally:
            self.repo.set_schema(origshema)

    def test_custom_attribute_param(self):
        origshema = self.repo.schema
        try:
            self.repo.config.repairing = True  # avoid versions checking
            self.repo.set_schema(self.repo.deserialize_schema())
            pes = self.repo.schema["Personne"]
            attr = pes.relation_definition("custom_field_of_jungle")
            self.assertIn("jungle_speed", vars(attr))
            self.assertEqual(42, attr.jungle_speed)
        finally:
            self.repo.set_schema(origshema)


class DataHelpersTC(CubicWebTC):
    def test_type_from_eid(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(self.repo.type_from_eid(2, cnx), "CWGroup")

    def test_type_from_eid_raise(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertRaises(UnknownEid, self.repo.type_from_eid, -2, cnx)

    def test_add_delete_info(self):
        with self.admin_access.repo_cnx() as cnx:
            entity = self.repo.vreg["etypes"].etype_class("Personne")(cnx)
            entity.eid = -1
            entity.complete = lambda x: None
            self.repo.add_info(cnx, entity, self.repo.system_source)
            cu = cnx.system_sql("SELECT * FROM entities WHERE eid = -1")
            data = cu.fetchall()
            self.assertEqual(tuplify(data), [(-1, "Personne")])
            self.repo._delete_cascade_multi(cnx, [entity])
            self.repo.system_source.delete_info_multi(cnx, [entity])
            cu = cnx.system_sql("SELECT * FROM entities WHERE eid = -1")
            data = cu.fetchall()
            self.assertEqual(data, [])


class FTITC(CubicWebTC):
    def test_fulltext_container_entity(self):
        with self.admin_access.repo_cnx() as cnx:
            assert (
                self.schema.relation_schema_for("use_email").fulltext_container
                == "subject"
            )
            toto = cnx.create_entity("EmailAddress", address="toto@logilab.fr")
            cnx.commit()
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "toto"})
            self.assertEqual(rset.rows, [])
            cnx.user.cw_set(use_email=toto)
            cnx.commit()
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "toto"})
            self.assertEqual(rset.rows, [[cnx.user.eid]])
            cnx.execute(
                'DELETE X use_email Y WHERE X login "admin", Y eid %(y)s',
                {"y": toto.eid},
            )
            cnx.commit()
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "toto"})
            self.assertEqual(rset.rows, [])
            tutu = cnx.create_entity("EmailAddress", address="tutu@logilab.fr")
            cnx.user.cw_set(use_email=tutu)
            cnx.commit()
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "tutu"})
            self.assertEqual(rset.rows, [[cnx.user.eid]])
            tutu.cw_set(address="hip@logilab.fr")
            cnx.commit()
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "tutu"})
            self.assertEqual(rset.rows, [])
            rset = cnx.execute("Any X WHERE X has_text %(t)s", {"t": "hip"})
            self.assertEqual(rset.rows, [[cnx.user.eid]])

    def test_no_uncessary_ftiindex_op(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity("Workflow", name="dummy workflow", description="huuuuu")
            self.assertFalse(
                any(
                    x
                    for x in cnx.pending_operations
                    if isinstance(x, native.FTIndexEntityOp)
                )
            )


class DBInitTC(CubicWebTC):
    def test_versions_inserted(self):
        with self.admin_access.repo_cnx() as cnx:
            inserted = [
                r[0]
                for r in cnx.execute(
                    "Any K ORDERBY K " 'WHERE P pkey K, P pkey ~= "system.version.%"'
                )
            ]
            self.assertEqual(
                inserted,
                [
                    "system.version.basket",
                    "system.version.card",
                    "system.version.comment",
                    "system.version.cubicweb",
                    "system.version.file",
                    "system.version.localperms",
                    "system.version.tag",
                ],
            )


CALLED = []


class InlineRelHooksTC(CubicWebTC):
    """test relation hooks are called for inlined relations"""

    def setUp(self):
        CubicWebTC.setUp(self)
        CALLED[:] = ()

    def test_inline_relation(self):
        """make sure <event>_relation hooks are called for inlined relation"""

        class EcritParHook(hook.Hook):
            __regid__ = "inlinedrelhook"
            __select__ = hook.Hook.__select__ & hook.match_rtype("ecrit_par")
            events = (
                "before_add_relation",
                "after_add_relation",
                "before_delete_relation",
                "after_delete_relation",
            )

            def __call__(self):
                CALLED.append((self.event, self.eidfrom, self.rtype, self.eidto))

        with self.temporary_appobjects(EcritParHook):
            with self.admin_access.repo_cnx() as cnx:
                eidp = cnx.execute('INSERT Personne X: X nom "toto"')[0][0]
                eidn = cnx.execute('INSERT Note X: X type "T"')[0][0]
                cnx.execute('SET N ecrit_par Y WHERE N type "T", Y nom "toto"')
                self.assertEqual(
                    CALLED,
                    [
                        ("before_add_relation", eidn, "ecrit_par", eidp),
                        ("after_add_relation", eidn, "ecrit_par", eidp),
                    ],
                )
                CALLED[:] = ()
                cnx.execute('DELETE N ecrit_par Y WHERE N type "T", Y nom "toto"')
                self.assertEqual(
                    CALLED,
                    [
                        ("before_delete_relation", eidn, "ecrit_par", eidp),
                        ("after_delete_relation", eidn, "ecrit_par", eidp),
                    ],
                )
                CALLED[:] = ()
                eidn = cnx.execute('INSERT Note N: N ecrit_par P WHERE P nom "toto"')[
                    0
                ][0]
                self.assertEqual(
                    CALLED,
                    [
                        ("before_add_relation", eidn, "ecrit_par", eidp),
                        ("after_add_relation", eidn, "ecrit_par", eidp),
                    ],
                )

    def test_unique_contraint(self):
        with self.admin_access.repo_cnx() as cnx:
            toto = cnx.create_entity("Personne", nom="toto")
            a01 = cnx.create_entity("Affaire", ref="A01", todo_by=toto)
            cnx.commit()
            cnx.create_entity("Note", type="todo", inline1=a01)
            cnx.commit()
            cnx.create_entity("Note", type="todo", inline1=a01)
            with self.assertRaises(ValidationError) as cm:
                cnx.commit()
            self.assertEqual(
                cm.exception.errors,
                {
                    "inline1-subject": (
                        "RQLUniqueConstraint S type T, S inline1 A1, "
                        "A1 todo_by C, Y type T, Y inline1 A2, A2 todo_by C failed"
                    )
                },
            )

    def test_add_relations_at_creation_with_del_existing_rel(self):
        with self.admin_access.repo_cnx() as cnx:
            person = cnx.create_entity(
                "Personne", nom="Toto", prenom="Lanturlu", sexe="M"
            )
            users_rql = 'Any U WHERE U is CWGroup, U name "users"'
            users = cnx.execute(users_rql).get_entity(0, 0)
            cnx.create_entity(
                "CWUser",
                login="Toto",
                upassword="firstname",
                firstname="firstname",
                surname="surname",
                reverse_login_user=person,
                in_group=users,
            )
            cnx.commit()


class PerformanceTest(CubicWebTC):
    def setUp(self):
        super().setUp()
        logger = logging.getLogger("cubicweb.session")
        # logger.handlers = [logging.StreamHandler(sys.stdout)]
        logger.setLevel(logging.INFO)
        self.info = logger.info

    def tearDown(self):
        super().tearDown()
        logger = logging.getLogger("cubicweb.session")
        logger.setLevel(logging.CRITICAL)

    def test_composite_deletion(self):
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            t0 = time.time()
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            abraham = cnx.create_entity(
                "Personne", nom="Abraham", prenom="John", sexe="M"
            )
            for j in range(0, 2000, 100):
                abraham.cw_set(personne_composite=personnes[j : j + 100])
            t1 = time.time()
            self.info("creation: %.2gs", (t1 - t0))
            cnx.commit()
            t2 = time.time()
            self.info("commit creation: %.2gs", (t2 - t1))
            cnx.execute("DELETE Personne P WHERE P eid %(eid)s", {"eid": abraham.eid})
            t3 = time.time()
            self.info("deletion: %.2gs", (t3 - t2))
            cnx.commit()
            t4 = time.time()
            self.info("commit deletion: %2gs", (t4 - t3))

    def test_add_relation_non_inlined(self):
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            cnx.commit()
            t0 = time.time()
            abraham = cnx.create_entity(
                "Personne",
                nom="Abraham",
                prenom="John",
                sexe="M",
                personne_composite=personnes[:100],
            )
            t1 = time.time()
            self.info("creation: %.2gs", (t1 - t0))
            for j in range(100, 2000, 100):
                abraham.cw_set(personne_composite=personnes[j : j + 100])
            t2 = time.time()
            self.info("more relations: %.2gs", (t2 - t1))
            cnx.commit()
            t3 = time.time()
            self.info("commit creation: %.2gs", (t3 - t2))

    def test_add_relation_inlined(self):
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            cnx.commit()
            t0 = time.time()
            abraham = cnx.create_entity(
                "Personne",
                nom="Abraham",
                prenom="John",
                sexe="M",
                personne_inlined=personnes[:100],
            )
            t1 = time.time()
            self.info("creation: %.2gs", (t1 - t0))
            for j in range(100, 2000, 100):
                abraham.cw_set(personne_inlined=personnes[j : j + 100])
            t2 = time.time()
            self.info("more relations: %.2gs", (t2 - t1))
            cnx.commit()
            t3 = time.time()
            self.info("commit creation: %.2gs", (t3 - t2))

    def test_session_add_relation(self):
        """to be compared with test_session_add_relations"""
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            abraham = cnx.create_entity(
                "Personne", nom="Abraham", prenom="John", sexe="M"
            )
            cnx.commit()
            t0 = time.time()
            add_relation = cnx.add_relation
            for p in personnes:
                add_relation(abraham.eid, "personne_composite", p.eid)
            cnx.commit()
            t1 = time.time()
            self.info("add relation: %.2gs", t1 - t0)

    def test_session_add_relations(self):
        """to be compared with test_session_add_relation"""
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            abraham = cnx.create_entity(
                "Personne", nom="Abraham", prenom="John", sexe="M"
            )
            cnx.commit()
            t0 = time.time()
            add_relations = cnx.add_relations
            relations = [
                ("personne_composite", [(abraham.eid, p.eid) for p in personnes])
            ]
            add_relations(relations)
            cnx.commit()
            t1 = time.time()
            self.info("add relations: %.2gs", t1 - t0)

    def test_session_add_relation_inlined(self):
        """to be compared with test_session_add_relations"""
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            abraham = cnx.create_entity(
                "Personne", nom="Abraham", prenom="John", sexe="M"
            )
            cnx.commit()
            t0 = time.time()
            add_relation = cnx.add_relation
            for p in personnes:
                add_relation(abraham.eid, "personne_inlined", p.eid)
            cnx.commit()
            t1 = time.time()
            self.info("add relation (inlined): %.2gs", t1 - t0)

    def test_session_add_relations_inlined(self):
        """to be compared with test_session_add_relation"""
        with self.admin_access.repo_cnx() as cnx:
            personnes = []
            for i in range(2000):
                p = cnx.create_entity(
                    "Personne", nom="Doe%03d" % i, prenom="John", sexe="M"
                )
                personnes.append(p)
            abraham = cnx.create_entity(
                "Personne", nom="Abraham", prenom="John", sexe="M"
            )
            cnx.commit()
            t0 = time.time()
            add_relations = cnx.add_relations
            relations = [
                ("personne_inlined", [(abraham.eid, p.eid) for p in personnes])
            ]
            add_relations(relations)
            cnx.commit()
            t1 = time.time()
            self.info("add relations (inlined): %.2gs", t1 - t0)

    def test_optional_relation_reset_1(self):
        with self.admin_access.repo_cnx() as cnx:
            p1 = cnx.create_entity("Personne", nom="Vincent")
            p2 = cnx.create_entity("Personne", nom="Florent")
            w = cnx.create_entity("Affaire", ref="wc")
            w.cw_set(todo_by=[p1, p2])
            w.cw_clear_all_caches()
            cnx.commit()
            self.assertEqual(len(w.todo_by), 1)
            self.assertEqual(w.todo_by[0].eid, p2.eid)

    def test_optional_relation_reset_2(self):
        with self.admin_access.repo_cnx() as cnx:
            p1 = cnx.create_entity("Personne", nom="Vincent")
            p2 = cnx.create_entity("Personne", nom="Florent")
            w = cnx.create_entity("Affaire", ref="wc")
            w.cw_set(todo_by=p1)
            cnx.commit()
            w.cw_set(todo_by=p2)
            w.cw_clear_all_caches()
            cnx.commit()
            self.assertEqual(len(w.todo_by), 1)
            self.assertEqual(w.todo_by[0].eid, p2.eid)


if __name__ == "__main__":
    unittest.main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""workflow handling:

* entity types defining workflow (Workflow, State, Transition...)
* workflow history (TrInfo)
* adapter for workflowable entities (IWorkflowableAdapter)
"""
from logilab.common.decorators import cached, clear_cache

from cubicweb import server
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import relation_possible


class WorkflowException(Exception):
    pass


class Workflow(AnyEntity):
    __regid__ = "Workflow"

    @property
    def initial(self):
        """return the initial state for this workflow"""
        return self.initial_state and self.initial_state[0] or None

    def is_default_workflow_of(self, etype):
        """return True if this workflow is the default workflow for the given
        entity type
        """
        return any(et for et in self.reverse_default_workflow if et.name == etype)

    def iter_workflows(self, _done=None):
        """return an iterator on actual workflows, eg this workflow and its
        subworkflows
        """
        # infinite loop safety belt
        if _done is None:
            _done = set()
        yield self
        _done.add(self.eid)
        for tr in self._cw.execute(
            "Any T WHERE T is WorkflowTransition, " "T transition_of WF, WF eid %(wf)s",
            {"wf": self.eid},
        ).entities():
            if tr.subwf.eid in _done:
                continue
            for subwf in tr.subwf.iter_workflows(_done):
                yield subwf

    # state / transitions accessors ############################################

    def state_by_name(self, statename):
        rset = self._cw.execute(
            "Any S, SN WHERE S name SN, S name %(n)s, " "S state_of WF, WF eid %(wf)s",
            {"n": statename, "wf": self.eid},
        )
        if rset:
            return rset.get_entity(0, 0)
        return None

    def state_by_eid(self, eid):
        rset = self._cw.execute(
            "Any S, SN WHERE S name SN, S eid %(s)s, " "S state_of WF, WF eid %(wf)s",
            {"s": eid, "wf": self.eid},
        )
        if rset:
            return rset.get_entity(0, 0)
        return None

    def transition_by_name(self, trname):
        rset = self._cw.execute(
            "Any T, TN WHERE T name TN, T name %(n)s, "
            "T transition_of WF, WF eid %(wf)s",
            {"n": trname, "wf": self.eid},
        )
        if rset:
            return rset.get_entity(0, 0)
        return None

    def transition_by_eid(self, eid):
        rset = self._cw.execute(
            "Any T, TN WHERE T name TN, T eid %(t)s, "
            "T transition_of WF, WF eid %(wf)s",
            {"t": eid, "wf": self.eid},
        )
        if rset:
            return rset.get_entity(0, 0)
        return None

    # wf construction methods ##################################################

    def add_state(self, name, initial=False, **kwargs):
        """add a state to this workflow"""
        state = self._cw.create_entity("State", name=name, **kwargs)
        self._cw.execute(
            "SET S state_of WF WHERE S eid %(s)s, WF eid %(wf)s",
            {"s": state.eid, "wf": self.eid},
        )
        if initial:
            assert not self.initial, f"Initial state already defined as {self.initial}"
            self._cw.execute(
                "SET WF initial_state S " "WHERE S eid %(s)s, WF eid %(wf)s",
                {"s": state.eid, "wf": self.eid},
            )
        return state

    def _add_transition(
        self, trtype, name, fromstates, requiredgroups=(), conditions=(), **kwargs
    ):
        tr = self._cw.create_entity(trtype, name=name, **kwargs)
        self._cw.execute(
            "SET T transition_of WF " "WHERE T eid %(t)s, WF eid %(wf)s",
            {"t": tr.eid, "wf": self.eid},
        )
        assert fromstates, fromstates
        if not isinstance(fromstates, (tuple, list)):
            fromstates = (fromstates,)
        for state in fromstates:
            if hasattr(state, "eid"):
                state = state.eid
            self._cw.execute(
                "SET S allowed_transition T " "WHERE S eid %(s)s, T eid %(t)s",
                {"s": state, "t": tr.eid},
            )
        tr.set_permissions(requiredgroups, conditions, reset=False)
        return tr

    def add_transition(
        self, name, fromstates, tostate=None, requiredgroups=(), conditions=(), **kwargs
    ):
        """add a transition to this workflow from some state(s) to another"""
        tr = self._add_transition(
            "Transition", name, fromstates, requiredgroups, conditions, **kwargs
        )
        if tostate is not None:
            if hasattr(tostate, "eid"):
                tostate = tostate.eid
            self._cw.execute(
                "SET T destination_state S " "WHERE S eid %(s)s, T eid %(t)s",
                {"t": tr.eid, "s": tostate},
            )
        return tr

    def add_wftransition(
        self,
        name,
        subworkflow,
        fromstates,
        exitpoints=(),
        requiredgroups=(),
        conditions=(),
        **kwargs,
    ):
        """add a workflow transition to this workflow"""
        tr = self._add_transition(
            "WorkflowTransition", name, fromstates, requiredgroups, conditions, **kwargs
        )
        if hasattr(subworkflow, "eid"):
            subworkflow = subworkflow.eid
        assert self._cw.execute(
            "SET T subworkflow WF WHERE WF eid %(wf)s,T eid %(t)s",
            {"t": tr.eid, "wf": subworkflow},
        )
        for fromstate, tostate in exitpoints:
            tr.add_exit_point(fromstate, tostate)
        return tr

    def replace_state(self, todelstate, replacement):
        """migration convenience method"""
        if not hasattr(todelstate, "eid"):
            todelstate = self.state_by_name(todelstate)
        if not hasattr(replacement, "eid"):
            replacement = self.state_by_name(replacement)
        args = {"os": todelstate.eid, "ns": replacement.eid}
        execute = self._cw.execute
        execute(
            "SET X in_state NS WHERE X in_state OS, " "NS eid %(ns)s, OS eid %(os)s",
            args,
        )
        execute(
            "SET X from_state NS WHERE X from_state OS, "
            "OS eid %(os)s, NS eid %(ns)s",
            args,
        )
        execute(
            "SET X to_state NS WHERE X to_state OS, " "OS eid %(os)s, NS eid %(ns)s",
            args,
        )
        todelstate.cw_delete()


class BaseTransition(AnyEntity):
    """customized class for abstract transition

    provides a specific may_be_fired method to check if the relation may be
    fired by the logged user
    """

    __regid__ = "BaseTransition"
    fetch_attrs, cw_fetch_order = fetch_config(["name", "type"])

    def __init__(self, *args, **kwargs):
        if self.cw_etype == "BaseTransition":
            raise WorkflowException("should not be instantiated")
        super(BaseTransition, self).__init__(*args, **kwargs)

    @property
    def workflow(self):
        return self.transition_of[0]

    def has_input_state(self, state):
        if hasattr(state, "eid"):
            state = state.eid
        return any(s for s in self.reverse_allowed_transition if s.eid == state)

    def may_be_fired(self, eid):
        """return true if the logged user may fire this transition

        `eid` is the eid of the object on which we may fire the transition
        """
        DBG = False
        if server.DEBUG & server.DBG_SEC:
            if "transition" in server._SECURITY_CAPS:
                DBG = True
        user = self._cw.user
        # check user is at least in one of the required groups if any
        groups = frozenset(g.name for g in self.require_group)
        if groups:
            matches = user.matching_groups(groups)
            if matches:
                if DBG:
                    print(
                        f"may_be_fired: {self.name!r} may fire: user matches {groups}"
                    )
                return matches
            if "owners" in groups and user.owns(eid):
                if DBG:
                    print(f"may_be_fired: {self.name!r} may fire: user is owner")
                return True
        # check one of the rql expression conditions matches if any
        if self.condition:
            if DBG:
                print(
                    "may_be_fired: %r: %s"
                    % (
                        self.name,
                        [
                            (
                                rqlexpr.expression,
                                rqlexpr.check_expression(self._cw, eid),
                            )
                            for rqlexpr in self.condition
                        ],
                    )
                )
            for rqlexpr in self.condition:
                if rqlexpr.check_expression(self._cw, eid):
                    return True
        if self.condition or groups:
            return False
        return True

    def set_permissions(self, requiredgroups=(), conditions=(), reset=True):
        """set or add (if `reset` is False) groups and conditions for this
        transition
        """
        if reset:
            self._cw.execute(
                "DELETE T require_group G WHERE T eid %(x)s", {"x": self.eid}
            )
            self._cw.execute("DELETE T condition R WHERE T eid %(x)s", {"x": self.eid})
        for gname in requiredgroups:
            rset = self._cw.execute(
                "SET T require_group G " "WHERE T eid %(x)s, G name %(gn)s",
                {"x": self.eid, "gn": gname},
            )
            assert rset, f"{gname} is not a known group"
        if isinstance(conditions, str):
            conditions = (conditions,)
        for expr in conditions:
            if isinstance(expr, str):
                kwargs = {"expr": expr}
            else:
                assert isinstance(expr, dict)
                kwargs = expr
            kwargs["x"] = self.eid
            kwargs.setdefault("mainvars", "X")
            self._cw.execute(
                'INSERT RQLExpression X: X exprtype "ERQLExpression", '
                "X expression %(expr)s, X mainvars %(mainvars)s, "
                "T condition X WHERE T eid %(x)s",
                kwargs,
            )
        # XXX clear caches?


class Transition(BaseTransition):
    """customized class for Transition entities"""

    __regid__ = "Transition"

    def dc_long_title(self):
        return f"{self.name} ({self._cw._(self.name)})"

    def destination(self, entity):
        try:
            return self.destination_state[0]
        except IndexError:
            return entity.cw_adapt_to("IWorkflowable").latest_trinfo().previous_state

    def potential_destinations(self):
        try:
            yield self.destination_state[0]
        except IndexError:
            for incomingstate in self.reverse_allowed_transition:
                for tr in incomingstate.reverse_destination_state:
                    for previousstate in tr.reverse_allowed_transition:
                        yield previousstate


class WorkflowTransition(BaseTransition):
    """customized class for WorkflowTransition entities"""

    __regid__ = "WorkflowTransition"

    @property
    def subwf(self):
        return self.subworkflow[0]

    def destination(self, entity):
        return self.subwf.initial

    def potential_destinations(self):
        yield self.subwf.initial

    def add_exit_point(self, fromstate, tostate):
        if hasattr(fromstate, "eid"):
            fromstate = fromstate.eid
        if tostate is None:
            self._cw.execute(
                "INSERT SubWorkflowExitPoint X: T subworkflow_exit X, "
                "X subworkflow_state FS WHERE T eid %(t)s, FS eid %(fs)s",
                {"t": self.eid, "fs": fromstate},
            )
        else:
            if hasattr(tostate, "eid"):
                tostate = tostate.eid
            self._cw.execute(
                "INSERT SubWorkflowExitPoint X: T subworkflow_exit X, "
                "X subworkflow_state FS, X destination_state TS "
                "WHERE T eid %(t)s, FS eid %(fs)s, TS eid %(ts)s",
                {"t": self.eid, "fs": fromstate, "ts": tostate},
            )

    def get_exit_point(self, entity, stateeid):
        """if state is an exit point, return its associated destination state"""
        if hasattr(stateeid, "eid"):
            stateeid = stateeid.eid
        try:
            tostateeid = self.exit_points()[stateeid]
        except KeyError:
            return None
        if tostateeid is None:
            # go back to state from which we've entered the subworkflow
            return (
                entity.cw_adapt_to("IWorkflowable")
                .subworkflow_input_trinfo()
                .previous_state
            )
        return self._cw.entity_from_eid(tostateeid)

    @cached
    def exit_points(self):
        result = {}
        for ep in self.subworkflow_exit:
            result[ep.subwf_state.eid] = ep.destination and ep.destination.eid
        return result

    def cw_clear_all_caches(self):
        super(WorkflowTransition, self).cw_clear_all_caches()
        clear_cache(self, "exit_points")


class SubWorkflowExitPoint(AnyEntity):
    """customized class for SubWorkflowExitPoint entities"""

    __regid__ = "SubWorkflowExitPoint"

    @property
    def subwf_state(self):
        return self.subworkflow_state[0]

    @property
    def destination(self):
        return self.destination_state and self.destination_state[0] or None


class State(AnyEntity):
    """customized class for State entities"""

    __regid__ = "State"
    fetch_attrs, cw_fetch_order = fetch_config(["name"])
    rest_attr = "eid"

    def dc_long_title(self):
        return f"{self.name} ({self._cw._(self.name)})"

    @property
    def workflow(self):
        # take care, may be missing in multi-sources configuration
        return self.state_of and self.state_of[0] or None


class TrInfo(AnyEntity):
    """customized class for Transition information entities"""

    __regid__ = "TrInfo"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["creation_date", "comment"], pclass=None
    )  # don't want modification_date

    @property
    def for_entity(self):
        return self.wf_info_for[0]

    @property
    def previous_state(self):
        return self.from_state[0]

    @property
    def new_state(self):
        return self.to_state[0]

    @property
    def transition(self):
        return self.by_transition and self.by_transition[0] or None


class IWorkflowableAdapter(EntityAdapter):
    """base adapter providing workflow helper methods for workflowable entities."""

    __regid__ = "IWorkflowable"
    __select__ = relation_possible("in_state")

    @cached
    def cwetype_workflow(self):
        """return the default workflow for entities of this type"""
        # XXX CWEType method
        wfrset = self._cw.execute(
            "Any WF WHERE ET default_workflow WF, ET name %(et)s",
            {"et": self.entity.cw_etype},
        )
        if wfrset:
            return wfrset.get_entity(0, 0)
        self.warning("can't find any workflow for %s", self.entity.cw_etype)
        return None

    @property
    def main_workflow(self):
        """return current workflow applied to this entity"""
        if self.entity.custom_workflow:
            return self.entity.custom_workflow[0]
        return self.cwetype_workflow()

    @property
    def current_workflow(self):
        """return current workflow applied to this entity"""
        return self.current_state and self.current_state.workflow or self.main_workflow

    @property
    def current_state(self):
        """return current state entity"""
        return self.entity.in_state and self.entity.in_state[0] or None

    @property
    def state(self):
        """return current state name"""
        try:
            return self.current_state.name
        except AttributeError:
            self.warning("entity %s has no state", self.entity)
            return None

    @property
    def printable_state(self):
        """return current state name translated to context's language"""
        state = self.current_state
        if state:
            return self._cw._(state.name)
        return ""

    @property
    def workflow_history(self):
        """return the workflow history for this entity (eg ordered list of
        TrInfo entities)
        """
        return self.entity.reverse_wf_info_for

    def latest_trinfo(self):
        """return the latest transition information for this entity"""
        try:
            return self.workflow_history[-1]
        except IndexError:
            return None

    def possible_transitions(self, type="normal"):
        """generates transition that MAY be fired for the given entity,
        expected to be in this state
        used only by the UI
        """
        if self.current_state is None or self.current_workflow is None:
            return
        rset = self._cw.execute(
            "Any T,TT, TN WHERE S allowed_transition T, S eid %(x)s, "
            "T type TT, T type %(type)s, "
            "T name TN, T transition_of WF, WF eid %(wfeid)s",
            {
                "x": self.current_state.eid,
                "type": type,
                "wfeid": self.current_workflow.eid,
            },
        )
        for tr in rset.entities():
            if tr.may_be_fired(self.entity.eid):
                yield tr

    def subworkflow_input_trinfo(self):
        """return the TrInfo which has be recorded when this entity went into
        the current sub-workflow
        """
        if self.main_workflow.eid == self.current_workflow.eid:
            return  # doesn't make sense
        subwfentries = []
        for trinfo in self.workflow_history:
            if (
                trinfo.transition
                and trinfo.previous_state.workflow.eid != trinfo.new_state.workflow.eid
            ):
                # entering or leaving a subworkflow
                if (
                    subwfentries
                    and subwfentries[-1].new_state.workflow.eid
                    == trinfo.previous_state.workflow.eid
                    and subwfentries[-1].previous_state.workflow.eid
                    == trinfo.new_state.workflow.eid
                ):
                    # leave
                    del subwfentries[-1]
                else:
                    # enter
                    subwfentries.append(trinfo)
        if not subwfentries:
            return None
        return subwfentries[-1]

    def subworkflow_input_transition(self):
        """return the transition which has went through the current sub-workflow"""
        return getattr(self.subworkflow_input_trinfo(), "transition", None)

    def _add_trinfo(self, comment, commentformat, treid=None, tseid=None):
        kwargs = {}
        if comment is not None:
            kwargs["comment"] = comment
            if commentformat is not None:
                kwargs["comment_format"] = commentformat
        kwargs["wf_info_for"] = self.entity
        if treid is not None:
            kwargs["by_transition"] = self._cw.entity_from_eid(treid)
        if tseid is not None:
            kwargs["to_state"] = self._cw.entity_from_eid(tseid)
        return self._cw.create_entity("TrInfo", **kwargs)

    def _get_transition(self, tr):
        assert self.current_workflow
        if isinstance(tr, str):
            _tr = self.current_workflow.transition_by_name(tr)
            assert _tr is not None, f"not a {self.__regid__} transition: {tr}"
            tr = _tr
        return tr

    def fire_transition(self, tr, comment=None, commentformat=None):
        """change the entity's state by firing given transition (name or entity)
        in entity's workflow
        """
        tr = self._get_transition(tr)
        return self._add_trinfo(comment, commentformat, tr.eid)

    def fire_transition_if_possible(self, tr, comment=None, commentformat=None):
        """change the entity's state by firing given transition (name or entity)
        in entity's workflow if this transition is possible
        """
        tr = self._get_transition(tr)
        if any(tr_ for tr_ in self.possible_transitions() if tr_.eid == tr.eid):
            self.fire_transition(tr, comment, commentformat)

    def change_state(self, statename, comment=None, commentformat=None, tr=None):
        """change the entity's state to the given state (name or entity) in
        entity's workflow. This method should only by used by manager to fix an
        entity's state when their is no matching transition, otherwise
        fire_transition should be used.
        """
        assert self.current_workflow
        if hasattr(statename, "eid"):
            stateeid = statename.eid
        else:
            state = self.current_workflow.state_by_name(statename)
            if state is None:
                raise WorkflowException(f"not a {self.__regid__} state: {statename}")
            stateeid = state.eid
        # XXX try to find matching transition?
        return self._add_trinfo(comment, commentformat, tr and tr.eid, stateeid)

    def set_initial_state(self, statename):
        """set a newly created entity's state to the given state (name or entity)
        in entity's workflow. This is useful if you don't want it to be the
        workflow's initial state.
        """
        assert self.current_workflow
        if hasattr(statename, "eid"):
            stateeid = statename.eid
        else:
            state = self.current_workflow.state_by_name(statename)
            if state is None:
                raise WorkflowException(f"not a {self.__regid__} state: {statename}")
            stateeid = state.eid
        self._cw.execute(
            "SET X in_state S WHERE X eid %(x)s, S eid %(s)s",
            {"x": self.entity.eid, "s": stateeid},
        )



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""schema for cubicweb.devtools tests"""
from yams.buildobjs import EntityType, SubjectRelation, String, RichString, Int, Date


class Person(EntityType):
    """a physical person"""

    surname = String(required=True, fulltextindexed=True, indexed=True, maxsize=64)
    firstname = String(fulltextindexed=True, maxsize=64)
    civility = String(
        required=True,
        internationalizable=True,
        vocabulary=("Mr", "Ms", "Mrs"),
        default="Mr",
    )
    description = RichString(fulltextindexed=True)
    birthday = Date()


class Bug(EntityType):
    title = String(maxsize=64, required=True, fulltextindexed=True)
    severity = String(vocabulary=("important", "normal", "minor"), default="normal")
    cost = Int()
    description = String(maxsize=4096, fulltextindexed=True)
    identical_to = SubjectRelation("Bug", symmetric=True)



# copyright 2019 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logging import getLogger

logger = getLogger("cubicweb")


SUBSCRIBERS = {
    "controller": [],
    "rql": [],
    "sql": [],
    "vreg": [],
    "registry_decisions": [],
}


def debug_channel_has_subscribers(channel):
    return len(SUBSCRIBERS.get(channel, [])) > 0


def subscribe_to_debug_channel(channel, subscriber):
    f"""
    Allow to subscribe a callable to one of the debug channels.

    The channel must be one of: {SUBSCRIBERS.keys()}

    And the callable need to accept one argument.

    It will raise Exception if the channel doesn't exist.
    """
    if channel not in SUBSCRIBERS.keys():
        raise Exception(f"debug channel '{channel}' doesn't exist")

    SUBSCRIBERS[channel].append(subscriber)


def unsubscribe_to_debug_channel(channel, subscriber):
    """
    Unsubscribe a callable from a channel. It will raise Exception if the
    channel doesn't exist nor
    """
    if channel not in SUBSCRIBERS.keys():
        raise Exception(f"debug channel '{channel}' doesn't exist")

    if subscriber not in SUBSCRIBERS[channel]:
        raise Exception(
            f"subscriber '{subscriber}' is not in debug channel '{channel}'"
        )

    SUBSCRIBERS[channel].remove(subscriber)


def emit_to_debug_channel(channel, message):
    """
    Send a message to a specified debug channel that will call all its
    subscribers.

    It will raise Exception if the channel doesn't exist.
    """
    if channel not in SUBSCRIBERS.keys():
        raise Exception(f"debug channel '{channel}' doesn't exist")

    for subscriber in SUBSCRIBERS[channel]:
        try:
            subscriber(message)
        except Exception:
            logger.error(
                "Failed to send debug message '%s' to subscriber '%s'",
                message,
                subscriber,
                exc_info=True,
            )



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittest for cubicweb user registration service"""

from yams import ValidationError
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb import Unauthorized


class RegisterUserTC(CubicWebTC):
    def test_register_user_service_anon(self):
        with self.new_access("anon").client_cnx() as cnx:
            self.assertRaises(
                Unauthorized,
                cnx.call_service,
                "register_user",
                login="foo2",
                password="bar2",
                email="foo2@bar2.com",
                firstname="Foo2",
                surname="Bar2",
            )

    def test_register_user_service_unique_login(self):
        with self.admin_access.cnx() as cnx:
            cnx.call_service(
                "register_user",
                login="foo3",
                password="bar3",
                email="foo3@bar3.com",
                firstname="Foo3",
                surname="Bar3",
            )
            # same login
            with self.assertRaises(ValidationError) as cm:
                cnx.call_service("register_user", login="foo3", password="bar3")
            expected_errors = {
                "": "some relations violate a unicity constraint",
                "login": "%(KEY-rtype)s is part of violated unicity constraint",
            }
            self.assertEqual(cm.exception.errors, expected_errors)

    def test_register_user_service_unique_email(self):
        with self.admin_access.cnx() as cnx:
            cnx.call_service(
                "register_user",
                login="foo3",
                password="bar3",
                email="foo3@bar3.com",
                firstname="Foo3",
                surname="Bar3",
            )
            with self.assertRaises(ValidationError) as cm:
                cnx.call_service(
                    "register_user", login="foo3@bar3.com", password="bar3"
                )
            expected_errors = {
                "": "some relations violate a unicity constraint",
                "address": "%(KEY-rtype)s is part of violated unicity constraint",
            }
            self.assertEqual(cm.exception.errors, expected_errors)

    def test_register_user_attributes(self):
        with self.admin_access.cnx() as cnx:
            cnx.call_service(
                "register_user",
                login="foo3",
                password="bar3",
                email="foo3@bar3.com",
                firstname="Foo3",
                surname="Bar3",
            )
            cnx.commit()

        with self.admin_access.client_cnx() as cnx:
            user = cnx.find("CWUser", login="foo3").one()
            self.assertEqual(user.firstname, "Foo3")
            self.assertEqual(user.use_email[0].address, "foo3@bar3.com")

    def test_register_user_groups(self):
        with self.admin_access.cnx() as cnx:
            # default
            cnx.call_service(
                "register_user",
                login="foo_user",
                password="bar_user",
                email="foo_user@bar_user.com",
                firstname="Foo_user",
                surname="Bar_user",
            )

            # group kwarg
            cnx.call_service(
                "register_user",
                login="foo_admin",
                password="bar_admin",
                email="foo_admin@bar_admin.com",
                firstname="Foo_admin",
                surname="Bar_admin",
                groups=("managers", "users"),
            )

            # class attribute
            from cubicweb.sobjects import services

            services.RegisterUserService.default_groups = ("guests",)
            cnx.call_service(
                "register_user",
                login="foo_guest",
                password="bar_guest",
                email="foo_guest@bar_guest.com",
                firstname="Foo_guest",
                surname="Bar_guest",
            )
            cnx.commit()

        with self.admin_access.client_cnx() as cnx:
            user = cnx.find("CWUser", login="foo_user").one()
            self.assertEqual([g.name for g in user.in_group], ["users"])

            admin = cnx.find("CWUser", login="foo_admin").one()
            self.assertEqual(
                sorted(g.name for g in admin.in_group), ["managers", "users"]
            )

            guest = cnx.find("CWUser", login="foo_guest").one()
            self.assertEqual([g.name for g in guest.in_group], ["guests"])


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import urljoin

from logilab.common.testlib import TestCase, unittest_main

from cubicweb import Unauthorized
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.req import RequestSessionAndConnectionBase


class RequestTC(TestCase):
    def test_rebuild_url(self):
        rebuild_url = RequestSessionAndConnectionBase(None).rebuild_url
        self.assertEqual(
            rebuild_url("http://logilab.fr?__message=pouet", __message="hop"),
            "http://logilab.fr?__message=hop",
        )
        self.assertEqual(
            rebuild_url("http://logilab.fr", __message="hop"),
            "http://logilab.fr?__message=hop",
        )
        self.assertEqual(
            rebuild_url("http://logilab.fr?vid=index", __message="hop"),
            "http://logilab.fr?__message=hop&vid=index",
        )

    def test_build_url(self):
        req = RequestSessionAndConnectionBase(None)
        req.from_controller = lambda: "view"
        req.relative_path = lambda includeparams=True: None
        req.base_url = lambda: BASE_URL
        self.assertEqual(req.build_url(), urljoin(BASE_URL, "view"))
        self.assertEqual(req.build_url(None), urljoin(BASE_URL, "view"))
        self.assertEqual(req.build_url("one"), urljoin(BASE_URL, "one"))
        self.assertEqual(req.build_url(param="ok"), urljoin(BASE_URL, "view?param=ok"))
        self.assertRaises(AssertionError, req.build_url, "one", "two not allowed")
        self.assertRaises(AssertionError, req.build_url, "view", test=None)

    def test_ensure_no_rql(self):
        req = RequestSessionAndConnectionBase(None)
        self.assertEqual(req.ensure_ro_rql("Any X WHERE X is CWUser"), None)
        self.assertEqual(req.ensure_ro_rql("  Any X WHERE X is CWUser  "), None)
        self.assertRaises(
            Unauthorized, req.ensure_ro_rql, 'SET X login "toto" WHERE X is CWUser'
        )
        self.assertRaises(
            Unauthorized,
            req.ensure_ro_rql,
            '   SET X login "toto" WHERE X is CWUser   ',
        )


class RequestCWTC(CubicWebTC):
    def test_base_url(self):
        base_url = self.config["base-url"]
        with self.admin_access.repo_cnx() as session:
            self.assertEqual(session.base_url(), base_url)

    def test_find(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="cdevienne",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
                in_group=cnx.find("CWGroup", name="users").one(),
            )

            cnx.create_entity(
                "CWUser",
                login="adim",
                upassword="adim",
                surname="di mascio",
                firstname="adrien",
                in_group=cnx.find("CWGroup", name="users").one(),
            )

            rset = cnx.find("CWUser", login="cdevienne")
            self.assertEqual(
                rset.printable_rql(), 'Any X WHERE X is CWUser, X login "cdevienne"'
            )
            u = rset.one()
            self.assertEqual(u.firstname, "Christophe")

            users = list(cnx.find("CWUser").entities())
            self.assertEqual(len(users), 4)

            groups = list(cnx.find("CWGroup", reverse_in_group=u).entities())
            self.assertEqual(len(groups), 1)
            self.assertEqual(groups[0].name, "users")

            rset = cnx.find("CWUser", in_group=groups[0])
            self.assertEqual(
                rset.printable_rql(),
                f"Any X WHERE X is CWUser, X in_group A, A eid {groups[0].eid}",
            )
            users = list(rset.entities())
            self.assertEqual(len(users), 2)

            with self.assertRaisesRegex(
                KeyError, "^'chapeau not in CWUser subject relations'$"
            ):
                cnx.find("CWUser", chapeau="melon")

            with self.assertRaisesRegex(
                KeyError, "^'buddy not in CWUser object relations'$"
            ):
                cnx.find("CWUser", reverse_buddy=users[0])

            with self.assertRaisesRegex(
                NotImplementedError,
                "^in_group: collection of values is not supported$",
            ):
                cnx.find("CWUser", in_group=[1, 2])

    def test_exists(self):
        with self.admin_access.cnx() as cnx:
            user_group = cnx.find("CWGroup", name="users").one()
            cnx.create_entity(
                "CWUser",
                login="rms",
                upassword="gnu",
                firstname="Richard",
                surname="Stallman",
                in_group=user_group,
            )

            self.assertFalse(cnx.exists("CWUser", login="not_in_the_db"))
            self.assertTrue(cnx.exists("CWUser", login="rms"))
            self.assertTrue(cnx.exists("CWUser", login="rms", in_group=user_group))


if __name__ == "__main__":
    unittest_main()



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from cubicweb.devtools import get_test_db_handler
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.repotest import BasePlannerTC, check_plan
from cubicweb.server.ssplanner import SSPlanner


def setUpModule(*args):
    # keep cnx so it's not garbage collected and the associated session closed
    global repo, cnx
    handler = get_test_db_handler(ApptestConfiguration("data", __file__))
    handler.build_db_cache()
    global repo, cnx
    repo, cnx = handler.get_repo_and_cnx()


def tearDownModule(*args):
    global repo, cnx
    del repo, cnx


class SSPlannerTC(BasePlannerTC):
    _test = check_plan

    def setUp(self):
        self.__class__.repo = repo
        BasePlannerTC.setUp(self)
        self.planner = SSPlanner(self.o.schema, self.repo.vreg.rqlhelper)
        self.system = self.o._repo.system_source

    def test_ordered_ambigous_sol(self):
        self._test(
            "Any XN ORDERBY XN WHERE X name XN, X is IN (Basket, State, Folder)",
            [
                (
                    "OneFetchStep",
                    [
                        (
                            "Any XN ORDERBY XN WHERE X name XN, "
                            "X is IN(Basket, State, Folder)",
                            [
                                {"X": "Basket", "XN": "String"},
                                {"X": "State", "XN": "String"},
                                {"X": "Folder", "XN": "String"},
                            ],
                        )
                    ],
                    [],
                )
            ],
        )

    def test_groupeded_ambigous_sol(self):
        self._test(
            "Any XN,COUNT(X) GROUPBY XN WHERE X name XN, X is IN (Basket, State, Folder)",
            [
                (
                    "OneFetchStep",
                    [
                        (
                            "Any XN,COUNT(X) GROUPBY XN WHERE X name XN, "
                            "X is IN(Basket, State, Folder)",
                            [
                                {"X": "Basket", "XN": "String"},
                                {"X": "State", "XN": "String"},
                                {"X": "Folder", "XN": "String"},
                            ],
                        )
                    ],
                    [],
                )
            ],
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



from cubicweb.hooks import notification
from cubicweb.predicates import is_instance


class FolderUpdateHook(notification.EntityUpdateHook):
    __select__ = notification.EntityUpdateHook.__select__ & is_instance("Folder")
    order = 100  # late trigger so that metadata hooks come before.



# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cw.server.migraction test"""
import datetime as dt

from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    SubjectRelation,
    Bytes,
    RichString,
    String,
    Int,
    Boolean,
    Datetime,
    Date,
    Float,
)
from yams.constraints import (
    SizeConstraint,
    UniqueConstraint,
    BoundaryConstraint,
    Attribute,
)

from cubicweb import _
from cubicweb.schema import (
    WorkflowableEntityType,
    RQLConstraint,
    RQLVocabularyConstraint,
    ERQLExpression,
    RRQLExpression,
)


class Affaire(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", ERQLExpression("X concerne S, S owned_by U")),
        "update": ("managers", "owners", ERQLExpression("X concerne S, S owned_by U")),
        "delete": ("managers", "owners", ERQLExpression("X concerne S, S owned_by U")),
    }

    ref = String(
        fulltextindexed=True, indexed=True, constraints=[SizeConstraint(max=16)]
    )
    sujet = String(fulltextindexed=True, constraints=[SizeConstraint(max=256)])
    concerne = SubjectRelation("Societe")
    opt_attr = Bytes()


class Societe(WorkflowableEntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "update": ("managers", "owners"),
        "delete": ("managers", "owners"),
        "add": (
            "managers",
            "users",
        ),
    }
    nom = String(maxsize=64, fulltextindexed=True)
    web = String(maxsize=128)
    tel = Float()
    fax = Int()
    rncs = String(maxsize=128)
    ad1 = String(maxsize=128)
    ad2 = String(maxsize=128)
    ad3 = String(maxsize=128)
    cp = String(maxsize=12)
    ville = String(maxsize=32)


# Division and SubDivision are gone


# New
class Para(EntityType):
    para = String(maxsize=512)
    newattr = String()
    newinlined = SubjectRelation("Affaire", cardinality="?*", inlined=True)
    newnotinlined = SubjectRelation("Affaire", cardinality="?*")


class Note(Para):
    __specializes_schema__ = True

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "update": (
            "managers",
            "owners",
        ),
        "delete": ("managers",),
        "add": (
            "managers",
            ERQLExpression(
                "X ecrit_part PE, U in_group G, "
                'PE require_permission P, P name "add_note", '
                "P require_group G"
            ),
        ),
    }

    whatever = Int(
        default=0
    )  # keep it before `date` for unittest_migraction.test_add_attribute_int
    yesno = Boolean(default=False)
    date = Datetime()
    type = String(maxsize=1)
    unique_id = String(maxsize=1, required=True, unique=True)
    mydate = Date(default="TODAY")
    newstyledefaultdate = Date(default=dt.date(2013, 1, 1))
    shortpara = String(
        maxsize=11, default="hop", vocabulary=["hop", "hop hop", "hop hop hop"]
    )
    ecrit_par = SubjectRelation(
        "Personne", constraints=[RQLConstraint("S concerne A, O concerne A")]
    )
    attachment = SubjectRelation("File")


class Frozable(EntityType):
    __permissions__ = {
        "read": ("managers", "users"),
        "add": ("managers", "users"),
        "update": (
            "managers",
            ERQLExpression("X frozen False"),
        ),
        "delete": (
            "managers",
            ERQLExpression("X frozen False"),
        ),
    }
    name = String()
    frozen = Boolean(
        default=False,
        __permissions__={
            "read": ("managers", "users"),
            "add": ("managers", "users"),
            "update": ("managers", "owners"),
        },
    )


class Personne(EntityType):
    __permissions__ = {
        "read": ("managers", "users"),  # 'guests' was removed
        "add": ("managers", "users"),
        "update": ("managers", "owners"),
        "delete": ("managers", "owners"),
    }
    __unique_together__ = [("nom", "prenom", "datenaiss")]
    nom = String(fulltextindexed=True, required=True, maxsize=64)
    prenom = String(fulltextindexed=True, maxsize=64)
    civility = String(maxsize=1, default="M", fulltextindexed=True)
    promo = String(vocabulary=("bon", "pasbon", "pasbondutout"))
    titre = String(fulltextindexed=True, maxsize=128)
    adel = String(maxsize=128)
    ass = String(maxsize=128)
    web = String(maxsize=128)
    tel = Int()
    fax = Int()
    datenaiss = Datetime()
    test = Boolean()

    travaille = SubjectRelation("Societe")
    concerne = SubjectRelation("Affaire")
    concerne2 = SubjectRelation(("Affaire", "Note"), cardinality="1*")
    connait = SubjectRelation("Personne", symmetric=True)


class concerne(RelationType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", RRQLExpression("U has_update_permission S")),
        "delete": ("managers", RRQLExpression("O owned_by U")),
    }


# `Old` entity type is gonce
# `comments` is gone
# `fiche` is gone
# `multisource_*` rdefs are gone
# `see_also_*` rdefs are gone


class evaluee(RelationDefinition):
    subject = ("Personne", "CWUser", "Societe")
    object = "Note"
    constraints = [RQLVocabularyConstraint("S owned_by U")]


class ecrit_par(RelationType):
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "delete": ("managers",),
        "add": (
            "managers",
            RRQLExpression(
                'O require_permission P, P name "add_note", '
                "U in_group G, P require_group G"
            ),
        ),
    }
    inlined = True
    cardinality = "?*"


# `copain` rdef is gone
# `tags` rdef is gone
# `filed_under` rdef is gone
# `require_permission` rdef is gone
# `require_state` rdef is gone
# `personne_composite` rdef is gone
# `personne_inlined` rdef is gone
# `login_user` rdef is gone
# `ambiguous_inlined` rdef is gone


class Folder(EntityType):
    """folders are used to classify entities. They may be defined as a tree."""

    name = String(required=True, indexed=True, internationalizable=True, maxsize=64)
    description = RichString(fulltextindexed=True)
    filed_under = SubjectRelation("Folder", description=_("parent folder"))


# New
class Text(Para):
    __specializes_schema__ = True
    summary = String(maxsize=512)


# New
class Folder2(EntityType):
    """folders are used to classify entities. They may be defined as a tree.
    When you include the Folder entity, all application specific entities
    may then be classified using the "filed_under" relation.
    """

    name = String(
        required=True,
        indexed=True,
        internationalizable=True,
        constraints=[UniqueConstraint(), SizeConstraint(max=64)],
    )
    description = RichString(fulltextindexed=True)


# New
class filed_under2(RelationDefinition):
    subject = "*"
    object = "Folder2"


# New
class New(EntityType):
    new_name = String(indexed=True, unique=True)


# New
class same_as(RelationDefinition):
    subject = ("Societe",)
    object = "ExternalUri"


class inlined_rel(RelationDefinition):
    subject = object = "Folder2"
    inlined = True
    cardinality = "??"


class Activity(EntityType):
    start = Datetime(
        constraints=[BoundaryConstraint("<=", Attribute("end"))],
        description=_("when the activity started"),
    )
    end = Datetime(
        constraints=[BoundaryConstraint(">=", Attribute("start"))],
        description=_("when the activity ended"),
    )



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.entities import AnyEntity, fetch_config, adapters
from cubicweb.predicates import is_instance


class Societe(AnyEntity):
    __regid__ = "Societe"
    fetch_attrs = ("nom",)


class Personne(Societe):
    """customized class forne Person entities"""

    __regid__ = "Personne"
    fetch_attrs, cw_fetch_order = fetch_config(["nom", "prenom"])
    rest_attr = "nom"


class Ami(Societe):
    __regid__ = "Ami"
    fetch_attrs, cw_fetch_order = fetch_config(["nom", "prenom"])
    rest_attr = "nom"


class Ville(AnyEntity):
    __regid__ = "Ville"
    rest_attr = "insee"


class Note(AnyEntity):
    __regid__ = "Note"


class FakeFileIDownloadableAdapter(adapters.IDownloadableAdapter):
    __select__ = is_instance("FakeFile")



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.cwconfig unit tests"""

import contextlib
import functools
import os
import pkgutil
import sys
import unittest
from os.path import dirname, join
from tempfile import TemporaryDirectory
from unittest.mock import patch

from logilab.common.modutils import cleanup_sys_modules
from pkg_resources import EntryPoint, Distribution

from cubicweb.cwconfig import CubicWebConfiguration, _expand_modname
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.testlib import BaseTestCase


def templibdir(func):
    """create a temporary directory and insert it in sys.path"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with TemporaryDirectory() as libdir:
            sys.path.insert(0, libdir)
            try:
                args = args + (libdir,)
                return func(*args, **kwargs)
            finally:
                sys.path.remove(libdir)

    return wrapper


def create_filepath(filepath):
    filedir = dirname(filepath)
    if not os.path.exists(filedir):
        os.makedirs(filedir)
    with open(filepath, "a"):
        pass


@contextlib.contextmanager
def temp_config(appid, instance_dir, cubes):
    """context manager that create a config object with specified appid,
    instance_dir and cubes"""
    cls = CubicWebConfiguration
    old = (cls._INSTANCES_DIR, sys.path[:], sys.meta_path[:])
    old_modules = set(sys.modules)
    try:
        cls._INSTANCES_DIR = instance_dir
        config = cls(appid)
        config._cubes = cubes
        config.adjust_sys_path()
        yield config
    finally:
        (cls._INSTANCES_DIR, sys.path[:], sys.meta_path[:]) = old
        for module in set(sys.modules) - old_modules:
            del sys.modules[module]


def iter_entry_points(group):
    """Mock importlib.metadata.entry_points to yield EntryPoint from
    packages found in test/data even though these are not
    installed.
    """
    libpython = CubicWebConfigurationTC.datapath()
    prefix = "cubicweb_"
    for pkgname in os.listdir(libpython):
        if not pkgname.startswith(prefix):
            continue
        location = join(libpython, pkgname)
        yield EntryPoint(pkgname[len(prefix) :], pkgname, dist=Distribution(location))


class CubicWebConfigurationTC(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        sys.path.append(cls.datapath())

    @classmethod
    def tearDownClass(cls):
        sys.path.remove(cls.datapath())

    def setUp(self):
        self.config = ApptestConfiguration("data", __file__)
        self.config._cubes = ("email", "file")

    def tearDown(self):
        cleanup_sys_modules([self.datapath()])

    def test_migration_scripts_dir(self):
        mscripts = os.listdir(self.config.migration_scripts_dir())
        self.assertIn("bootstrapmigration_repository.py", mscripts)
        self.assertIn("postcreate.py", mscripts)
        self.assertIn("3.24.0_Any.py", mscripts)

    @patch(
        (
            "importlib_metadata.entry_points"
            if sys.version_info < (3, 10)
            else "importlib.metadata.entry_points"
        ),
        side_effect=iter_entry_points,
    )
    def test_available_cubes(self, mock_iter_entry_points):
        expected_cubes = [
            "cubicweb_card",
            "cubicweb_comment",
            "cubicweb_email",
            "cubicweb_file",
            "cubicweb_forge",
            "cubicweb_localperms",
            "cubicweb_mycube",
            "cubicweb_required_variables",
            "cubicweb_tag",
        ]
        self.assertEqual(self.config.available_cubes(), expected_cubes)
        mock_iter_entry_points.assert_called_once_with(group="cubicweb.cubes")

    def test_reorder_cubes(self):
        # forge depends on email and file and comment
        # email depends on file
        self.assertEqual(
            self.config.reorder_cubes(["file", "email", "forge"]),
            ("forge", "email", "file"),
        )
        self.assertEqual(
            self.config.reorder_cubes(["email", "file", "forge"]),
            ("forge", "email", "file"),
        )
        self.assertEqual(
            self.config.reorder_cubes(["email", "forge", "file"]),
            ("forge", "email", "file"),
        )
        self.assertEqual(
            self.config.reorder_cubes(["file", "forge", "email"]),
            ("forge", "email", "file"),
        )
        self.assertEqual(
            self.config.reorder_cubes(["forge", "file", "email"]),
            ("forge", "email", "file"),
        )
        self.assertEqual(
            self.config.reorder_cubes(("forge", "email", "file")),
            ("forge", "email", "file"),
        )

    def test_reorder_cubes_recommends(self):
        from cubicweb_comment import __pkginfo__ as comment_pkginfo

        self._test_reorder_cubes_recommends(comment_pkginfo)

    def _test_reorder_cubes_recommends(self, comment_pkginfo):
        comment_pkginfo.__recommends_cubes__ = {"file": None}
        try:
            # email recommends comment
            # comment recommends file
            self.assertEqual(
                self.config.reorder_cubes(("forge", "email", "file", "comment")),
                ("forge", "email", "comment", "file"),
            )
            self.assertEqual(
                self.config.reorder_cubes(("forge", "email", "comment", "file")),
                ("forge", "email", "comment", "file"),
            )
            self.assertEqual(
                self.config.reorder_cubes(("forge", "comment", "email", "file")),
                ("forge", "email", "comment", "file"),
            )
            self.assertEqual(
                self.config.reorder_cubes(("comment", "forge", "email", "file")),
                ("forge", "email", "comment", "file"),
            )
        finally:
            comment_pkginfo.__recommends_cubes__ = {}

    def test_expand_cubes(self):
        self.assertEqual(
            self.config.expand_cubes(("email", "comment")),
            ["email", "comment", "file"],
        )

    def test_init_cubes_ignore_pyramid_cube(self):
        warning_msg = "cubicweb-pyramid got integrated into CubicWeb"
        with self.assertLogs("cubicweb.configuration", level="WARNING") as cm:
            self.config.init_cubes(["pyramid", "card"])
        self.assertIn(warning_msg, cm.output[0])
        self.assertNotIn("pyramid", self.config._cubes)

    @patch(
        (
            "importlib_metadata.entry_points"
            if sys.version_info < (3, 10)
            else "importlib.metadata.entry_points"
        ),
        side_effect=iter_entry_points,
    )
    def test_ccplugin_modname(self, mock_iter_entry_points):
        self.config.load_cwctl_plugins()
        mock_iter_entry_points.assert_called_once_with(group="cubicweb.cubes")
        self.assertIn("cubicweb_mycube.ccplugin", sys.modules, sorted(sys.modules))

    def test_config_value_from_environment_str(self):
        self.assertIsNone(self.config["base-url"])
        os.environ["CW_BASE_URL"] = "https://www.cubicweb.org"
        try:
            self.assertEqual(self.config["base-url"], "https://www.cubicweb.org")
        finally:
            del os.environ["CW_BASE_URL"]

    def test_config_value_from_environment_int(self):
        self.assertEqual(self.config["connections-pool-max-size"], 0)
        os.environ["CW_CONNECTIONS_POOL_MAX_SIZE"] = "6"
        try:
            self.assertEqual(self.config["connections-pool-max-size"], 6)
        finally:
            del os.environ["CW_CONNECTIONS_POOL_MAX_SIZE"]

    def test_config_value_from_environment_yn(self):
        self.assertEqual(self.config["allow-email-login"], False)
        try:
            for val, result in (
                ("yes", True),
                ("no", False),
                ("y", True),
                ("n", False),
            ):
                os.environ["CW_ALLOW_EMAIL_LOGIN"] = val
            self.assertEqual(self.config["allow-email-login"], result)
        finally:
            del os.environ["CW_ALLOW_EMAIL_LOGIN"]

    def test_config_value_from_environment_required(self):
        # check that without any variable environment, required options are not
        # loaded.
        config = ApptestConfiguration("data", __file__)
        config._cubes = ("required_variables",)
        config.load_site_cubicweb()

        self.assertNotIn("a-required-string-option", config)
        self.assertNotIn("a-required-int-option", config)
        self.assertNotIn("a-required-yn-option", config)

        # insert the options, and reload the configuration
        os.environ["CW_A_REQUIRED_STRING_OPTION"] = "42"
        os.environ["CW_A_REQUIRED_INT_OPTION"] = "42"
        os.environ["CW_A_REQUIRED_YN_OPTION"] = "y"

        try:
            config = ApptestConfiguration("data", __file__)
            config._cubes = ("required_variables",)
            config.load_site_cubicweb()

            self.assertEqual(config["a-required-string-option"], "42")
            self.assertEqual(config["a-required-int-option"], 42)
            self.assertEqual(config["a-required-yn-option"], True)
        finally:
            del os.environ["CW_A_REQUIRED_STRING_OPTION"]
            del os.environ["CW_A_REQUIRED_INT_OPTION"]
            del os.environ["CW_A_REQUIRED_YN_OPTION"]


class ModnamesTC(unittest.TestCase):
    @templibdir
    def test_expand_modnames(self, libdir):
        tempdir = join(libdir, "lib")
        filepaths = [
            join(tempdir, "__init__.py"),
            join(tempdir, "a.py"),
            join(tempdir, "b.py"),
            join(tempdir, "c.py"),
            join(tempdir, "b", "__init__.py"),
            join(tempdir, "b", "a.py"),
            join(tempdir, "b", "c.py"),
            join(tempdir, "b", "d", "__init__.py"),
            join(tempdir, "e", "e.py"),
        ]
        for filepath in filepaths:
            create_filepath(filepath)
        # not importable
        self.assertEqual(list(_expand_modname("isnotimportable")), [])
        # not a python package
        self.assertEqual(list(_expand_modname("lib.e")), [])
        self.assertEqual(
            list(_expand_modname("lib.a")),
            [
                ("lib.a", join(tempdir, "a.py")),
            ],
        )
        # lib.b.d should be imported
        self.assertEqual(
            list(_expand_modname("lib.b")),
            [
                ("lib.b", join(tempdir, "b", "__init__.py")),
                ("lib.b.a", join(tempdir, "b", "a.py")),
                ("lib.b.c", join(tempdir, "b", "c.py")),
                ("lib.b.d", join(tempdir, "b", "d", "__init__.py")),
            ],
        )
        # lib.b.d should not be imported without recursive mode
        self.assertEqual(
            list(_expand_modname("lib.b", recursive=False)),
            [
                ("lib.b", join(tempdir, "b", "__init__.py")),
                ("lib.b.a", join(tempdir, "b", "a.py")),
                ("lib.b.c", join(tempdir, "b", "c.py")),
            ],
        )
        self.assertEqual(
            list(_expand_modname("lib")),
            [
                ("lib", join(tempdir, "__init__.py")),
                ("lib.a", join(tempdir, "a.py")),
                ("lib.b", join(tempdir, "b", "__init__.py")),
                ("lib.b.a", join(tempdir, "b", "a.py")),
                ("lib.b.c", join(tempdir, "b", "c.py")),
                ("lib.b.d", join(tempdir, "b", "d", "__init__.py")),
                ("lib.c", join(tempdir, "c.py")),
            ],
        )
        for source in (
            join(tempdir, "c.py"),
            join(tempdir, "b", "c.py"),
            join(tempdir, "b", "d", "__init__.py"),
        ):
            # remove source file
            os.remove(source)
        self.assertEqual(list(_expand_modname("lib.c")), [])
        self.assertEqual(
            list(_expand_modname("lib.b")),
            [
                ("lib.b", join(tempdir, "b", "__init__.py")),
                ("lib.b.a", join(tempdir, "b", "a.py")),
            ],
        )
        self.assertEqual(
            list(_expand_modname("lib")),
            [
                ("lib", join(tempdir, "__init__.py")),
                ("lib.a", join(tempdir, "a.py")),
                ("lib.b", join(tempdir, "b", "__init__.py")),
                ("lib.b.a", join(tempdir, "b", "a.py")),
            ],
        )

    @templibdir
    def test_schema_modnames(self, libdir):
        for filepath in (
            join(libdir, "schema.py"),
            join(libdir, "cubicweb_foo", "__init__.py"),
            join(libdir, "cubicweb_foo", "schema", "__init__.py"),
            join(libdir, "cubicweb_foo", "schema", "a.py"),
            join(libdir, "cubicweb_foo", "schema", "b.py"),
            # subpackages should not be loaded
            join(libdir, "cubicweb_foo", "schema", "c", "__init__.py"),
            join(libdir, "_instance_dir", "data1", "schema.py"),
            join(libdir, "_instance_dir", "data2", "noschema.py"),
        ):
            create_filepath(filepath)
        expected = [
            ("cubicweb", "cubicweb.schemas.bootstrap"),
            ("cubicweb", "cubicweb.schemas.base"),
            ("cubicweb", "cubicweb.schemas.workflow"),
            ("cubicweb", "cubicweb.schemas.Bookmark"),
            ("foo", "cubicweb_foo.schema"),
            ("foo", "cubicweb_foo.schema.a"),
            ("foo", "cubicweb_foo.schema.b"),
        ]
        # app has schema file
        instance_dir = join(libdir, "_instance_dir")
        with temp_config("data1", instance_dir, ("foo",)) as config:
            self.assertEqual(
                pkgutil.find_loader("schema").get_filename(),
                join(libdir, "_instance_dir", "data1", "schema.py"),
            )
            self.assertEqual(config.schema_modnames(), expected + [("data", "schema")])
        # app doesn't have schema file
        with temp_config("data2", instance_dir, ("foo",)) as config:
            self.assertEqual(
                pkgutil.find_loader("schema").get_filename(), join(libdir, "schema.py")
            )
            self.assertEqual(config.schema_modnames(), expected)

    @templibdir
    def test_appobjects_modnames(self, libdir):
        for filepath in (
            join(libdir, "entities.py"),
            join(libdir, "cubicweb_foo", "__init__.py"),
            join(libdir, "cubicweb_foo", "entities", "__init__.py"),
            join(libdir, "cubicweb_foo", "entities", "a.py"),
            # subpackages should be loaded recursively
            join(libdir, "cubicweb_foo", "entities", "b", "__init__.py"),
            join(libdir, "cubicweb_foo", "entities", "b", "a.py"),
            join(libdir, "cubicweb_foo", "entities", "b", "c", "__init__.py"),
            join(libdir, "cubicweb_foo", "hooks.py"),
            join(libdir, "_instance_dir", "data1", "entities.py"),
            join(libdir, "_instance_dir", "data2", "hooks.py"),
        ):
            create_filepath(filepath)
        instance_dir = join(libdir, "_instance_dir")
        expected = [
            "cubicweb.entities",
            "cubicweb.entities.adapters",
            "cubicweb.entities.authobjs",
            "cubicweb.entities.lib",
            "cubicweb.entities.schemaobjs",
            "cubicweb.entities.sources",
            "cubicweb.entities.wfobjs",
            "cubicweb.hooks",
            "cubicweb.hooks.bookmark",
            "cubicweb.hooks.email",
            "cubicweb.hooks.integrity",
            "cubicweb.hooks.metadata",
            "cubicweb.hooks.notification",
            "cubicweb.hooks.security",
            "cubicweb.hooks.synccomputed",
            "cubicweb.hooks.syncschema",
            "cubicweb.hooks.syncsession",
            "cubicweb.hooks.syncsources",
            "cubicweb.hooks.workflow",
            "cubicweb.sobjects",
            "cubicweb.sobjects.ldapparser",
            "cubicweb.sobjects.notification",
            "cubicweb.sobjects.services",
            "cubicweb.sobjects.supervising",
            "cubicweb_foo.entities",
            "cubicweb_foo.entities.a",
            "cubicweb_foo.entities.b",
            "cubicweb_foo.entities.b.a",
            "cubicweb_foo.entities.b.c",
            "cubicweb_foo.hooks",
        ]
        # data1 has entities
        with temp_config("data1", instance_dir, ("foo",)) as config:
            config.cube_appobject_path = {"entities", "hooks"}
            self.assertEqual(config.appobjects_modnames(), expected + ["entities"])
        # data2 has hooks
        with temp_config("data2", instance_dir, ("foo",)) as config:
            config.cube_appobject_path = {"entities", "hooks"}
            self.assertEqual(config.appobjects_modnames(), expected + ["hooks"])


if __name__ == "__main__":
    unittest.main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""core CubicWeb schema, but not necessary at bootstrap time"""


from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    SubjectRelation,
    String,
    Bytes,
    TZDatetime,
    Password,
)

from cubicweb import _
from cubicweb.schema import (
    RQLConstraint,
    WorkflowableEntityType,
    ERQLExpression,
    RRQLExpression,
    PUB_SYSTEM_REL_PERMS,
    PUB_SYSTEM_ATTR_PERMS,
    RO_ATTR_PERMS,
)


class CWUser(WorkflowableEntityType):
    """define a CubicWeb user"""

    __permissions__ = {
        "read": ("managers", "users", ERQLExpression("X identity U")),
        "add": ("managers",),
        "delete": ("managers",),
        "update": (
            "managers",
            ERQLExpression('X identity U, NOT U in_group G, G name "guests"'),
        ),
    }

    login = String(
        required=True,
        unique=True,
        maxsize=64,
        description=_("unique identifier used to connect to the application"),
    )
    upassword = Password(required=True)  # password is a reserved word for mysql
    firstname = String(maxsize=64)
    surname = String(maxsize=64)
    last_login_time = TZDatetime(description=_("last connection date"))
    in_group = SubjectRelation(
        "CWGroup",
        cardinality="+*",
        constraints=[RQLConstraint('NOT O name "owners"')],
        description=_("groups grant permissions to the user"),
    )


class EmailAddress(EntityType):
    """an electronic mail address associated to a short alias"""

    __permissions__ = {
        # application that wishes public email, or use it for something else
        # than users (eg Company, Person), should explicitly change permissions
        "read": ("managers", ERQLExpression("EXISTS(U use_email X)")),
        "add": (
            "managers",
            "users",
        ),
        "delete": (
            "managers",
            "owners",
            ERQLExpression("EXISTS(P use_email X, U has_update_permission P)"),
        ),
        "update": (
            "managers",
            "owners",
            ERQLExpression("EXISTS(P use_email X, U has_update_permission P)"),
        ),
    }

    alias = String(fulltextindexed=True, maxsize=56)
    address = String(
        required=True, fulltextindexed=True, indexed=True, unique=True, maxsize=128
    )
    prefered_form = SubjectRelation(
        "EmailAddress",
        cardinality="?*",
        description=_(
            "when multiple addresses are equivalent \
(such as python-projects@logilab.org and python-projects@lists.logilab.org), set this \
to indicate which is the preferred form."
        ),
    )


class use_email(RelationType):
    fulltext_container = "subject"


class use_email_relation(RelationDefinition):
    """user's email account"""

    name = "use_email"
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
        "delete": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
    }
    subject = "CWUser"
    object = "EmailAddress"
    cardinality = "*?"
    composite = "subject"


class primary_email(RelationDefinition):
    """the prefered email"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
        "delete": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
    }
    subject = "CWUser"
    object = "EmailAddress"
    cardinality = "??"
    constraints = [RQLConstraint("S use_email O")]


class prefered_form(RelationType):
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        # XXX should have update __permissions__ on both subject and object,
        #     though by doing this we will probably have no way to add
        #     this relation in the web ui. The easiest way to acheive this
        #     is probably to be able to have "U has_update_permission O" as
        #     RQLConstraint of the relation definition, though this is not yet
        #     possible
        "add": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
        "delete": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
    }


class in_group(RelationType):
    """core relation indicating a user's groups"""

    __permissions__ = PUB_SYSTEM_REL_PERMS


class owned_by(RelationType):
    """core relation indicating owners of an entity. This relation
    implicitly put the owner into the owners group for the entity
    """

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (
            "managers",
            RRQLExpression("S owned_by U"),
        ),
        "delete": (
            "managers",
            RRQLExpression("S owned_by U"),
        ),
    }
    # 0..n cardinality for entities created by internal session (no attached user)
    # and to support later deletion of a user which has created some entities
    cardinality = "**"
    subject = "*"
    object = "CWUser"


class created_by(RelationType):
    """core relation indicating the original creator of an entity"""

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    # 0..1 cardinality for entities created by internal session (no attached user)
    # and to support later deletion of a user which has created some entities
    cardinality = "?*"
    subject = "*"
    object = "CWUser"


class creation_date(RelationType):
    """creation time of an entity"""

    __permissions__ = PUB_SYSTEM_ATTR_PERMS
    cardinality = "11"
    subject = "*"
    object = "TZDatetime"


class modification_date(RelationType):
    """latest modification time of an entity"""

    __permissions__ = PUB_SYSTEM_ATTR_PERMS
    cardinality = "11"
    subject = "*"
    object = "TZDatetime"


class cwuri(RelationType):
    """internal entity uri"""

    __permissions__ = RO_ATTR_PERMS
    cardinality = "11"
    subject = "*"
    object = "String"


# XXX find a better relation name
class for_user(RelationType):
    """link a property to the user which want this property customization. Unless
    you're a site manager, this relation will be handled automatically.
    """

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    inlined = True
    subject = "CWProperty"
    object = "CWUser"
    composite = "object"
    cardinality = "?*"


class ExternalUri(EntityType):
    """a URI representing an object in external data store"""

    uri = String(
        required=True, unique=True, maxsize=256, description=_("the URI of the object")
    )


class same_as(RelationType):
    """generic relation to specify that an external entity represent the same
    object as a local one:
       http://www.w3.org/TR/owl-ref/#sameAs-def
    """

    # NOTE: You'll have to explicitly declare which entity types can have a
    # same_as relation
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": ("managers", "users"),
        "delete": ("managers", "owners"),
    }
    cardinality = "**"
    symmetric = True
    # NOTE: the 'object = ExternalUri' declaration will still be mandatory
    #       in the cube's schema.
    object = "ExternalUri"


class CWSource(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "update": ("managers",),
        "delete": ("managers",),
    }
    name = String(
        required=True, unique=True, maxsize=128, description=_("name of the source")
    )
    type = String(required=True, maxsize=20, description=_("type of the source"))
    config = String(
        description=_(
            "source's configuration. One key=value per "
            "line, authorized keys depending on the "
            "source's type"
        ),
        __permissions__={
            "read": ("managers",),
            "add": ("managers",),
            "update": ("managers",),
        },
    )
    # put this here and not in a subclass even if it's only for some sources
    # since having subclasses on generic relation (cw_source) double the number
    # of rdef in the schema, and make ms planning harder since queries solutions
    # may changes when sources are specified
    url = String(
        description=_(
            "URLs from which content will be imported. You can put one url per line"
        )
    )
    parser = String(
        description=_(
            "parser to use to extract entities from content retrieved at given URLs."
        )
    )
    latest_retrieval = TZDatetime(description=_("latest synchronization time"))
    in_synchronization = TZDatetime(
        description=_(
            "start timestamp of the currently in synchronization, or NULL when no synchronization in progress."
        )
    )


ENTITY_MANAGERS_PERMISSIONS = {
    "read": ("managers",),
    "add": ("managers",),
    "update": ("managers",),
    "delete": ("managers",),
}
RELATION_MANAGERS_PERMISSIONS = {
    "read": ("managers",),
    "add": ("managers",),
    "delete": ("managers",),
}


class CWSourceHostConfig(EntityType):
    __permissions__ = ENTITY_MANAGERS_PERMISSIONS
    __unique_together__ = [("match_host", "cw_host_config_of")]
    match_host = String(
        required=True,
        maxsize=128,
        description=_("regexp matching host(s) to which this config applies"),
    )
    config = String(
        required=True,
        description=_(
            "Source's configuration for a particular host. "
            "One key=value per line, authorized keys "
            "depending on the source's type, overriding "
            "values defined on the source."
        ),
        __permissions__={
            "read": ("managers",),
            "add": ("managers",),
            "update": ("managers",),
        },
    )


class cw_host_config_of(RelationDefinition):
    __permissions__ = RELATION_MANAGERS_PERMISSIONS
    subject = "CWSourceHostConfig"
    object = "CWSource"
    cardinality = "1*"
    composite = "object"
    inlined = True


class cw_source(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    subject = "*"
    object = "CWSource"
    cardinality = "1*"
    composite = "object"


class CWDataImport(EntityType):
    __permissions__ = ENTITY_MANAGERS_PERMISSIONS
    start_timestamp = TZDatetime()
    end_timestamp = TZDatetime()
    log = String()
    status = String(
        required=True,
        internationalizable=True,
        indexed=True,
        default="in progress",
        vocabulary=[_("in progress"), _("success"), _("failed")],
    )


class cw_import_of(RelationDefinition):
    __permissions__ = RELATION_MANAGERS_PERMISSIONS
    subject = "CWDataImport"
    object = "CWSource"
    cardinality = "1*"
    composite = "object"


# "abtract" relation types, no definition in cubicweb itself ###################


class identical_to(RelationType):
    """identical to"""

    symmetric = True
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        # XXX should have update __permissions__ on both subject and object,
        #     though by doing this we will probably have no way to add
        #     this relation in the web ui. The easiest way to acheive this
        #     is probably to be able to have "U has_update_permission O" as
        #     RQLConstraint of the relation definition, though this is not yet
        #     possible
        "add": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
        "delete": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
    }


class see_also(RelationType):
    """generic relation to link one entity to another"""

    symmetric = True
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
        "delete": (
            "managers",
            RRQLExpression("U has_update_permission S"),
        ),
    }


class CWSession(EntityType):
    """Persistent session.

    Used by cubicweb.pyramid to store the session data.
    """

    __permissions__ = {
        "read": ("managers",),
        "add": (),
        "update": (),
        "delete": (),
    }
    cwsessiondata = Bytes()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Some i18n/gettext utilities."""
import os
from glob import glob
from os.path import join, basename, splitext, exists

from cubicweb.toolsutils import create_dir


def add_msg(w, msgid, msgctx=None):
    """write an empty pot msgid definition"""
    if msgctx:
        w(f'msgctxt "{msgctx}"\n')
    msgid = msgid.replace('"', r"\"").splitlines()
    if len(msgid) > 1:
        w('msgid ""\n')
        for line in msgid:
            w('"%s"' % line.replace('"', r"\""))
    else:
        w(f'msgid "{msgid[0]}"\n')
    w('msgstr ""\n\n')


def execute2(args):
    # XXX replace this with check_output in Python 2.7
    from subprocess import Popen, PIPE, CalledProcessError

    p = Popen(args, stdout=PIPE, stderr=PIPE)
    out, err = p.communicate()
    if p.returncode != 0:
        exc = CalledProcessError(p.returncode, args[0])
        exc.cmd = args
        exc.data = (out, err)
        raise exc


def available_catalogs(i18ndir=None):
    if i18ndir is None:
        wildcard = "*.po"
    else:
        wildcard = join(i18ndir, "*.po")
    for popath in glob(wildcard):
        lang = splitext(basename(popath))[0]
        yield lang, popath


def compile_i18n_catalogs(sourcedirs, destdir, langs):
    """generate .mo files for a set of languages into the `destdir` i18n directory"""
    from subprocess import CalledProcessError
    from logilab.common.fileutils import ensure_fs_mode

    print(f"-> compiling message catalogs to {destdir}")
    errors = []
    for lang in langs:
        langdir = join(destdir, lang, "LC_MESSAGES")
        if not exists(langdir):
            create_dir(langdir)
        pofiles = [join(path, f"{lang}.po") for path in sourcedirs]
        pofiles = [pof for pof in pofiles if exists(pof)]
        mergedpo = join(destdir, f"{lang}_merged.po")
        try:
            # merge instance/cubes messages catalogs with the stdlib's one
            cmd = [
                "msgcat",
                "--use-first",
                "--sort-output",
                "--strict",
                "-o",
                mergedpo,
            ] + pofiles
            execute2(cmd)
            # make sure the .mo file is writeable and compiles with *msgfmt*
            applmo = join(destdir, lang, "LC_MESSAGES", "cubicweb.mo")
            try:
                ensure_fs_mode(applmo)
            except OSError:
                pass  # suppose not exists
            execute2(["msgfmt", mergedpo, "-o", applmo])
        except CalledProcessError as exc:
            errors.append(
                "while handling language %s:\ncmd:\n%s\nstdout:\n%s\nstderr:\n%s\n"
                % (lang, exc.cmd, repr(exc.data[0]), repr(exc.data[1]))
            )
        except Exception as exc:
            errors.append(f"while handling language {lang}: {exc}")
        try:
            # clean everything
            os.unlink(mergedpo)
        except Exception:
            continue
    return errors
